#!/usr/bin/perl -w
#
#
#
# procedure for looking at frame shift info in BLASTX data
# requires full BLASTX report, not parsed ones
#
# WORKFLOW
# open file and grab first report
# for first hit
# grab first HSP data: strand, frame, query start and stop, hit start and stop
# for each next HSP, if same strand, different frame, compare coordinates:
#       query must have slight overhang or be separate,
#       and hit cannot be within coords seen in previous hit
#
# OUTPUT columns...
# 1   query name
# 2   query length
# 3   hit name
# 4   hit length
# 5   hsp1 query strand
# 6   hsp1 query frame
# 7   hsp+1 query strand
# 8   hsp+1 query frame
# 9   hsp1 query start
# 10  hsp1 query end
# 11  hsp+1 query start
# 12  hsp+1 query end
# 13  hsp1 hit start
# 14  hsp1 hit end
# 15  hsp+1 hit start
# 16  hsp+1 hit end
# 17  hsp1 E value
# 18  hsp1 % ID
# 19  hsp1 bit score
# 20  hsp+1 E value
# 21  hsp+1 % ID
# 22  hsp+1 bit score
# 23  hsp counter
#
# CHANGES
# 10/21/13, rmf
#   updated usage instructions
#   added proper command line switches and switch processing
#   fixed output so automatically goes to output file instead of STDOUT
#
#
# 3/XX/14, rmf
#   changed to accept XML input
#
# 4/4/13
#   BLAST parser does not seem to like the new BLAST executables and their
#    output format. Need to grab sequence name from description line. Result->query_name is broken




use strict;
use Bio::SearchIO;

my $usage 							= qq (
detect_fs_by_BLASTX.pl -i infile [-o outfilebase] [-h]

Will read in standard BLASTX full text report data and will output likely candidates for 
frameshift correction, including coordinates of change on both query and target.

	-i      input BLASTX report filename
	-o      output filename     [default: inputfile + ".fs_detect.data"]
	-x      BLAST file is in XML format
			
	-h      this help text

);

# set defaults and parse args...
my ($infile, $datfile);
my $datfile_suffix      = '.fs_detect.dat';
my $blastformat = 'blast';

# and parse args
parse_args();


our $logFile;
our $LOGFH;
my $debug = 0;


if ($debug) {
    my $logname = TimeStamp();
    $logname =~ s/ /_/g;
    $logname =~ s/://g;
    open ($LOGFH, ">log_" . $logname) || die "Error opening log file log_$logname for writing: $!\n";
}

open (my $DATFH, ">$datfile") || die "Error opening output data file $datfile for writing: $!\n";
my $result_fh   = Bio::SearchIO->new ('-file'     => "<$infile",
                                      '-format'   => $blastformat);
my $counter = 0;

while ( my $result = $result_fh->next_result() )  {
    #
    print STDERR "Processing BLAST report #" . "$counter...\r" if $counter++ % 1000 == 0;
    
    next if $result->num_hits() == 0;

    # grab 1st hit and important stats
    my $hit1 = $result->next_hit();
    
    # skip all the rest of this if there's only 1 hsp
    next if $hit1->num_hsps() == 1;
    
    my $hsp1 = $hit1->next_hsp();
    my $hsp_counter = 1;
    
    my $hsp1_strand      = $hsp1->strand('query');
    my $hsp1_frame       = $hsp1->query->frame();
    my $hsp1_query_start = $hsp1->start('query');
    my $hsp1_query_end   = $hsp1->end('query');
    my $hsp1_hit_start   = $hsp1->start('hit');
    my $hsp1_hit_end     = $hsp1->end('hit');

    while ( my $next_hsp = $hit1->next_hsp() ) {
        #
        $hsp_counter++;
        if ($next_hsp->strand('query') == $hsp1_strand && $next_hsp->query->frame() != $hsp1_frame) {
            # basic two conditions met, where strand is the same but frame is different
            # now check coordinates
            my $next_hsp_query_start = $next_hsp->start('query');
            my $next_hsp_query_end   = $next_hsp->end('query');
            my $next_hsp_hit_start   = $next_hsp->start('hit');
            my $next_hsp_hit_end     = $next_hsp->end('hit');
            
            # DEBUG
            if ($debug) {
                print $LOGFH join("\t", $hsp1_query_start,$hsp1_query_end,$next_hsp_query_start,$next_hsp_query_end,$result->query_name(),
                                        $hsp1_hit_start,$hsp1_hit_end,$next_hsp_hit_start,$next_hsp_hit_end,$hit1->name() )
                             . "\n";
            }
            
            
            # switch pairs if 2nd pair comes before 1st one
            my $query_coord_switch = 0;
            if ($next_hsp_query_start < $hsp1_query_start) {
                ($hsp1_query_start, $hsp1_query_end, $next_hsp_query_start, $next_hsp_query_end)
                    = ($next_hsp_query_start, $next_hsp_query_end, $hsp1_query_start, $hsp1_query_end);
                $query_coord_switch = 1;
            }; 
            # now check overlap
            my $query_overlap   = describe_overlap( $hsp1_query_start, $hsp1_query_end,
                                                    $next_hsp_query_start, $next_hsp_query_end, $result->query_name());
            # switch back if needed
            ($next_hsp_query_start, $next_hsp_query_end, $hsp1_query_start, $hsp1_query_end)
                = ($hsp1_query_start, $hsp1_query_end, $next_hsp_query_start, $next_hsp_query_end)
                if $query_coord_switch;


            # switch pairs if 2nd pair comes before 1st one
            my $hit_coord_switch = 0;
            if ($next_hsp_hit_start < $hsp1_hit_start) {
                ($hsp1_hit_start, $hsp1_hit_end, $next_hsp_hit_start, $next_hsp_hit_end)
                  = ($next_hsp_hit_start, $next_hsp_hit_end, $hsp1_hit_start, $hsp1_hit_end);
                $hit_coord_switch = 1;
            }
            # now check overlap
            my $hit_overlap     = describe_overlap( $hsp1_hit_start, $hsp1_hit_end,
                                                    $next_hsp_hit_start, $next_hsp_hit_end, $hit1->name());
            # switch back if needed
            ($next_hsp_hit_start, $next_hsp_hit_end, $hsp1_hit_start, $hsp1_hit_end)
                = ($hsp1_hit_start, $hsp1_hit_end, $next_hsp_hit_start, $next_hsp_hit_end)
                if $hit_coord_switch;
            
            
            if (    ($query_overlap eq 'none' or $query_overlap eq 'right_overlap')
                and ($hit_overlap eq 'none' or $query_overlap eq 'right_overlap')) {
                # got a good pair, so let's output all the info...
                
                # CHANGE: rmf, 4/4/14
                #print {$DATFH}  $result->query_name(), "\t";
                my ($temp_query_name) = $result->query_description() =~ m/^(\S+)/;
                print {$DATFH}  $temp_query_name, "\t";
                # END CHANGE
                
                print {$DATFH}  $result->query_length(), "\t";
                print {$DATFH}  $hit1->name(), "\t";
                print {$DATFH}  $hit1->length(), "\t";
                
                print {$DATFH}  $hsp1_strand, "\t";
                print {$DATFH}  $hsp1_frame, "\t";
                print {$DATFH}  $next_hsp->strand('query'), "\t";
                print {$DATFH}  $next_hsp->query->frame(), "\t";
                
                print {$DATFH}  $hsp1_query_start, "\t";
                print {$DATFH}  $hsp1_query_end, "\t";
                print {$DATFH}  $next_hsp_query_start, "\t";
                print {$DATFH}  $next_hsp_query_end, "\t";
                
                print {$DATFH}  $hsp1_hit_start, "\t";
                print {$DATFH}  $hsp1_hit_end, "\t";
                print {$DATFH}  $next_hsp_hit_start, "\t";
                print {$DATFH}  $next_hsp_hit_end, "\t";
                
                print {$DATFH}  $hsp1->evalue(), "\t";
                print {$DATFH}  $hsp1->percent_identity(), "\t";
                print {$DATFH}  $hsp1->bits(), "\t";
                
                print {$DATFH}  $next_hsp->evalue(), "\t";
                print {$DATFH}  $next_hsp->percent_identity(), "\t";
                print {$DATFH}  $next_hsp->bits(), "\t";
                
                print {$DATFH}  $hsp_counter, "\n";
                
            }
        
        }
    
    
    
    } # end $next_hsp = ...
} # end $report = ...


if ($debug) {
    close $LOGFH;
}

close $DATFH;
print STDERR "\n";


exit;

#
#
# END OF PROGRAM
#

sub parse_args {
	# parse arguments
	while (scalar(@ARGV)) {
		my $arg						= shift @ARGV ;
		if 	  ($arg eq '-h') 		{ die $usage;                    } 
		elsif ($arg eq '-i') 		{ $infile 		= shift @ARGV;   }
		elsif ($arg eq '-o') 		{ $datfile	 	= shift @ARGV;   }
		elsif ($arg eq '-x')        { $blastformat  = 'blastxml'     }
		else 						{ die "unknown argument '$arg'"; }
	}
	
	# -i input file
	if (!defined $infile) {
		print "You must supply an input filename using '-i' argument!\n";
		die $usage;
	};
	
	# -o output file basename
	if (!defined $datfile) {
	    $datfile = $infile . $datfile_suffix;
	}

}


sub describe_overlap {
    #
    # requires that first set preceeds second set
    my ($a_left, $a_right, $b_left, $b_right, $source) = @_;
    
    # DEBUG
    #print $LOGFH join("\t", $a_left,$a_right,$b_left,$b_right,$source) . "\n";
    
    # check for no overlap -- the easiest...
    if ($a_left < $b_left and $a_right < $b_left) {
        return 'none';
    } elsif ($a_left < $b_left and $a_right >= $b_left and $a_right < $b_right) {
        return 'right_overlap';
    } elsif ($a_left > $b_left and $a_left <= $b_left and $a_right > $b_right) {
        return 'left_overlap';
    } else {
        return 'overlap';
    }
}

sub Log {
	#return if (not $debugFlag);
	my $timestamp = TimeStamp();
	if ($logFile) {
		# write to a real log file
		open(L,">>$logFile");
		print L "$timestamp: ".$_[0];
		close L;
	} else {	
		# write to STDERR
		print STDERR "$timestamp: ".$_[0];
	}
}

sub TimeStamp {
  my @mytime = localtime(time);
  my $yr = 1900 + $mytime[5];
  my $mo = ZeroPad(1 + $mytime[4]);
  my $da = ZeroPad($mytime[3]);
  my $hr = ZeroPad($mytime[2]);
  my $mi = ZeroPad($mytime[1]);
  my $se = ZeroPad($mytime[0]);
  return "$yr$mo$da $hr:$mi:$se";
}

sub ZeroPad {
  return ($_[0]<10?"0$_[0]":$_[0]);
}

sub TimeWord {
  my @mytime = localtime(time);
  my $yr = 1900 + $mytime[5];
  my $mo = ("Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec")[$mytime[4]];
  my $wday = ("Sun","Mon","Tue","Wed","Thu","Fri","Sat")[$mytime[6]];
  my $da = $mytime[3];
  my $hr = $mytime[2];
  my $mi = $mytime[1];
  my $se = $mytime[0];
  return "$wday, $mo $da, $yr at $hr:$mi:$se";
}

