#!/usr/bin/perl -w


# rewrite_fasta_by_frame.pl in_fasta frame_data
#
# 

# CHANGES
#   11/8/13, rmf
#       was changing value of iterator in translate_files; fixed
#       forgot to specify alphabet & display_ID of translated frame0 objects; fixed
#       forgot to remove any trailing '*', as CD-HIT will complain; fixed
#       if trimming trims back to nothing, would return nothing; now discards seq
#       processing translations now will skip input seqobj if length is zero


use Bio::SeqIO;
use Carp;
use Data::Dumper;
use English	qw ( -no_match_vars );


my $usage 							= qq (
	ms_translate_fasta_by_frame.pl -i in_nt_fasta -o out_pep_fasta -d frame_data_file

	takes frame data from a parsed BLAST results file and translates the input file
	in fasta format into a fasta peptide output file
);


# open input files and output file
# parse hit_data data file into hash
# step thru each seq in fasta file
# 	if in hash, output to proper translation frame file
#	if not in hash, write out to no frame file
# next
# translate files
# concatenate

$OUTPUT_AUTOFLUSH = 1;
our %frame_data;
our %frame_file;
our %frame_fh;
our @frames;

our ( $infile, $outfile, $datafile );
parse_args();
parse_frame_data();
setup_frame_files();
separate_fasta_by_frame();
translate_files();
process_to_final_file();
cleanup();

exit;


#
#
# END OF PROGRAM
#

sub parse_args {
	#
	while (scalar(@ARGV)) {
		my $arg						= shift(@ARGV);
		if	  ($arg eq '-h') 		{ die    	$usage; } 
		elsif ($arg eq '-i') 		{ $infile 	= shift(@ARGV); }
		elsif ($arg eq '-o') 		{ $outfile 	= shift(@ARGV); }
		elsif ($arg eq '-d')		{ $datafile	= shift(@ARGV); }
		else 						{ die 		"unknown argument '$arg'"; }
	}

	# check to ensure we have the required parameters
	if (!defined $infile or !defined $outfile or !defined $datafile) {
		print "Error from undefined input arguments!\n";
		die $usage;
	}
	
}

sub parse_frame_data {
	# data file from xGI folks. columns will be parsed as:
	# 0 SEQ_NAME		grab this!
	# 1 HIT_NAME
	# 2 %ID
	# 3 %SIM
	# 4 SEQ_LEFT		grab this!
	# 5 SEQ_RIGHT		grab this!
	# 6 HIT_LEFT
	# 7 HIT_RIGHT
	# 8 DESCRIPTION
	# 9 HIT_LEN|COV_LEN
	#10 SCORE
	
    print STDERR "Parsing translation table file...\n";
	open (my $DATAFILE, "<$datafile")	|| die "Error opening file $datafile for reading: $!\n";

	while ( my $line = <$DATAFILE> ) {
		chomp $line;
		next if $line =~ m/^\s|#/;
		
		# this line could have been written as a slice, but we'll keep it simple now
		my ($name, $frame)				= split /\t/, $line;
		$frame_data{$name}			= $frame;
	}
	close $DATAFILE;
}

sub setup_frame_files {
	
	@frames = (-3, -2, -1, 0, 1, 2, 3);
	foreach my $item (@frames) {
		$frame_file{$item} = 'frame' . $item . '.fasta';
	}
}

sub separate_fasta_by_frame {

    print STDERR "Creating frame files...\n";
    
	# create input SeqIO object
	my $seq_in = Bio::SeqIO->new(	'-file'   => "<$infile",
                             		'-format' => 'fasta' );
	# create output SeqIO objects
	foreach my $item (@frames) {
		$frame_fh{$item} = Bio::SeqIO->new(	'-file'   => ">".$frame_file{$item},
                             				'-format' => 'fasta' );	
	}
	
	# grab next sequence
	while (my $inseq = $seq_in->next_seq()) {
	
	# see if it exists in hash; if so, write to appropriate frame
		#  if not, write to frame 0 file
		my $primary_id = $inseq->primary_id();
		if (defined $frame_data{$primary_id}) {
			$frame = $frame_data{$primary_id};
			$frame_fh{$frame}->write_seq($inseq);
			
		} else {
			$frame_fh{0}->write_seq($inseq);
		}
	} #end of while
	
}

sub translate_files2 {
    #
    print STDERR "Translating...\n";
    
    # read in each sequence file and do straight translation ....
    
	foreach my $item (@frames) {
	    #next if $item == 0;
		my $fasta = $frame_file{$item};
		my $frame_phrase = ($item == 0) ? 'all' : $item;
		my $command = "/groups/acornworm/opt/dna2pep/dna2pep.py -r $frame_phrase -x $fasta --fasta $fasta.pep";
		print STDERR "$command\n";
		`$command`;
		
	}


}


sub translate_files {

    print STDERR "Translating...\n";
    
	# /groups/acornworm/opt/dna2pep/dna2pep.py -o none -r all input_40/Trinity.fasta.part10 --fasta part10.pep
	foreach my $frame_phase (@frames) {
	
	    #next if $frame_phase == 0;
		my $fasta  = $frame_file{$frame_phase};
		
		my $seqin  = Bio::SeqIO->new( -file => $fasta,          -format => 'fasta');
		my $seqout = Bio::SeqIO->new( -file => ">$fasta.pep",   -format => 'fasta');

        if ($frame_phase != 0) {
		    while ( my $seqobj = $seqin->next_seq() ) {
		        # traslate and write out
		        next if $seqobj->length() == 0;
		        # CHANGE, RMF, 11/8/13
		        # bad! should not change value of iterator!
		        my $frame_phase_adj = $frame_phase;
		        #
		        if ($frame_phase < 0) {
		            $seqobj = $seqobj->revcom();
		            # CHANGE, RMF, 11/8/13
		            # bad! should not change value of iterator!
		            $frame_phase_adj = abs ($frame_phase);
		            #$frame_phase = abs ($frame_phase);
		            #
		        }
		        $seqout->write_seq($seqobj->translate( -frame => $frame_phase_adj - 1 ));
            }
        }
        else {
            while ( my $seqobj = $seqin->next_seq() ) {
		        # traslate longest ORF and write out
		        next if $seqobj->length() == 0;
                my $sequence = $seqobj->seq();
                my $notstrict = 0;
                my ($length, $start, $end, $direction, $orf_sequence) = longestORF($sequence, $notstrict);
                # CHANGE, RMF 11/8/13
                # forgot to specify alphabet of new object...
                my $orf = Bio::Seq->new( -seq => $orf_sequence, -alphabet => 'dna', -display_id => $seqobj->display_id() );
                #
                my $translated_orf = $orf->translate();

		        $seqout->write_seq($translated_orf);
		    }

        }
		
	}
}

sub process_to_final_file {

    print STDERR "Processing translations...\n";
    
    # open each frame file except 0
    # read each sequence
    # subsitute stops and process longest peptide
    my $seq_out = Bio::SeqIO->new(	'-file'   => ">$outfile",
                             		'-format' => 'fasta' );	
    
    foreach my $item (@frames) {
        #
        #print STDERR "$item\n";
        open (my $FRAMEDEBUG, ">" . $frame_file{$item} . ".debug")
            || die "Error open frame debug file '$item' for writing: $!\n";
        #next if $item == 0;
        
	    $frame_fh{$item} = Bio::SeqIO->new(	'-file'   => "<" . $frame_file{$item} . ".pep",
                             				'-format' => 'fasta' );	
        
        while (my $inseq = $frame_fh{$item}->next_seq()) {
            # do it all in one fell swoop
            
            # CHANGE, 11/8/13, rmf
            #   verify that we actually have something of length to work with!!
            next if $inseq->length() == 0;
            #
            my ($longest,$tempseq) = trim_longest_from_peptide( $inseq->seq(), $inseq->display_id() );
            # CHANGE, 11/8/13, rmf
            #   will skip to next seq if this one is completely trimmed back
            next if length ($tempseq) == 0;
            #
            print $FRAMEDEBUG join ("\t", $inseq->display_id(),$item,$longest,$tempseq), "\n";
            
            if (length ($tempseq) > 6 ) {
                $inseq->seq($tempseq);
                # and write it out...
                $seq_out->write_seq($inseq) if $item != 0;
            }
            # if < 25, just drop the sequence
        } #while
        
        close $FRAMEDEBUG;
        
    } #foreach

}

sub trim_longest_from_peptide {
    #
    # with each translation, split at the stops and grab the longest peptide
    my ($sequence, $display_id) = @_;
    $sequence =~ s/\*/\*J\*/g;
    my @peptide = split /J/, $sequence;
    #my $longest = (reverse sort { $a <=> $b } map { length($_) } @peptide)[0];
    #                @articles = sort {$b <=> $a} @files;  # sort numerically descending
    my $longest = (sort { length $b <=> length $a } @peptide)[0];
    
    # and now trim back        
    if (substr ($longest,0,1) eq '*' && substr ($longest,-1,1) eq '*') {
        # * at the start and the end
        # this is an internal longest peptide. Trim back only at the front to the first M
        # IMPROVE: need to check start coord of hit to see if at N-term of protein
        $sequence = trim_front_to_M($longest);
    }
    elsif (substr ($longest,0,1) eq '*') {
        # * at the start only
        # we have a peptide from the end of the ORF; trim front to M and back to k
        # IMPROVE: need to check start coord of hit to see if at N-term of protein
        $sequence = trim_front_to_M($longest);
        $sequence = trim_end_to_K($sequence);
    }
    elsif (substr ($longest,-1,1) eq '*') {
        # * at the end only
        # we have a peptide from the start of the translation; trim front to K
        # IMPROVE: ???
        $sequence = trim_front_to_K($longest);
    }
    else {
        # * not found
        #print STDERR "Full open reading frame on $display_id! Trimming anyhow...\n";
        # we have an internal peptide; trim front to K and end to K
        $sequence = trim_front_to_K($longest);
        $sequence = trim_end_to_K($sequence);
    }
    # CHANGE 11/8/13, RMF
    #  need to remove '*' so that CD-HIT doesn't barf
    if (length ($sequence) > 0) {
        $sequence =~ s|\*||g;
    }
    #
    return ($longest, $sequence);
}

sub trim_front_to_M {
    my $seq = shift;
    $seq =~ s/^.+?M/M/;
    return $seq;
}

sub trim_front_to_K {
    my $seq = shift;
    # CHANGE: rmf, 3/28/13
    # added R to regex
    # orig: $seq =~ s/^.+?K/K/;
    $seq =~ s/^.+?(K|R)/$1/;
    # END CHANGE
    return $seq;
}

sub trim_end_to_K {
    my $seq = shift;
    #$seq =~ s/K.+?$/K/;
    # CHANGE: rmf, 3/28/13
    # added R to regex
    # orig: $seq =~ m/^(.+K)/;
    $seq =~ m/^(.+K|.+R)/;
    # END CHANGE
    if (defined $1) {
        return $1;
    }
    else {
        return '*';
    }
}

sub cleanup {
    `mkdir frame_files` if (! -e 'frame_files');
    @frames = (-3, -2, -1, 0, 1, 2, 3);
	foreach my $item (@frames) {
	    my $frame_glob = 'frame' . $item . '.*';
	    `mv $frame_glob frame_files/`;
	}
}

sub longestORF {
   my $best=0;
   my ($bests,$beste,$beststrand)=(-1,-1,0);
   my $bestorf="";

   my $relaxed=$_[1];
   my $dna=Bio::Seq->new(-seq => $_[0]);
   my %strand=('+'=>$dna->seq,
               '-'=>$dna->revcom->seq);

   foreach my $direction (keys %strand) {
      my @starts=();
      my @ends=();
      if ($relaxed) {
         for (my $frame=0;$frame<3;$frame++) {
            unless ($strand{$direction}=~m/^.{$frame}(taa|tga|tag)/i) {
               push @starts,$frame+1;
            }
         }
      }
      while ($strand{$direction}=~m/(atg)/gi) {
         push @starts,pos($strand{$direction})-2;
      }

      while ($strand{$direction}=~m/(taa|tga|tag)/gi) {
         push @ends,pos($strand{$direction})-2;
      }
      push @ends,($dna->length-2,$dna->length-1,$dna->length);

      for my $s (@starts) {
         for my $e (@ends) {
            if ($e%3==$s%3 and $e>$s) {
               if ($e-$s>$best) {
                  $best=$e-$s;
                  ($bests,$beste,$beststrand)=($s,$e,$direction);
                  $bestorf=Bio::Seq->new(-seq=>$strand{$direction})->subseq($s,$e);
               }
               last
            } else {
               next
            }
         }
      }
   }
   return ($best,$bests,$beste,$beststrand,$bestorf);
}

